defmodule OASIS.Line do
  defstruct predicted_previous: 0,
    predicted_next: 0,
    original_line_number: 0,
    original_line: "",
    prediction_tree: []

  def parse({line, line_number}) do
    prediction_tree =
      line
      |> String.trim()
      |> String.split()
      |> Enum.map(fn ds -> Integer.parse(ds) |> elem(0) end)
      |> predict()


    %OASIS.Line{
      original_line_number: line_number,
      predicted_previous: expected_previous(prediction_tree),
      predicted_next: expected_next(prediction_tree),
      original_line: String.trim(line),
      prediction_tree: prediction_tree
    }
  end

  defp predict(line_numbers) do
    line_numbers
    |> prediction_tree()
  end

  defp expected_previous(prediction_tree) do
    prediction_tree
    |> List.first
    |> List.first
  end

  defp expected_next(prediction_tree) do
    prediction_tree
    |> List.first
    |> List.last
  end

  # Takes the list of numbers making up a singgle line of the OASIS result, and recursively calculate the prediction tree for it.
  # For example: (Note this example is from part1, we're now calculating in both directions in a single pass.)
  # [1,2,3] -> [[1,2,3,4],[1,1,1],[0,0]]
  defp prediction_tree(this_line_arr) do
    if Enum.uniq(this_line_arr) == [0] do
      # This is the final case, if we've gotten to the end then we add a new zero to the list and start spooling back upwards.
      [[0] ++ this_line_arr ++ [0]]
    else
      diff_line = 
        0..(Enum.count(this_line_arr)-2) # We're 1-indexed, and we only want to go to the 2nd-to-last item.
        |> Enum.reduce([], fn i, acc ->
          acc ++ [Enum.at(this_line_arr, i+1) - Enum.at(this_line_arr, i)]
        end)
      next_line = prediction_tree(diff_line)
      prev_diff = List.first(this_line_arr) - List.first(List.first(next_line))
      next_diff = List.last(this_line_arr) + List.last(List.first(next_line))
      [[prev_diff] ++ this_line_arr ++ [next_diff] | next_line]
    end
  end
end

defmodule OASIS do
  defstruct lines: [], sum_of_next_predictions: 0, sum_of_prev_predictions: 0

  def parse_file(filename) do
    File.read!(filename)
    |> String.trim()
    |> String.split("\n")
    |> Enum.with_index()
    |> Enum.map(&Task.async(OASIS.Line, :parse, [&1]))
    |> Enum.map(&Task.await/1)
    |> then(fn lines -> 
      %OASIS{
        lines: lines,
        sum_of_prev_predictions: Enum.reduce(lines, 0, fn l, acc -> acc+l.predicted_previous end),
        sum_of_next_predictions: Enum.reduce(lines, 0, fn l, acc -> acc+l.predicted_next end),
      }
    end)
  end
end


IO.puts("Considering the OASIS results...")
o = OASIS.parse_file("input")

IO.puts(" - Sanity Checks....")
IO.puts("   - A basic all-positive-number line with the same step the whole way gets the correct prev prediction: #{Enum.find(o.lines, fn l -> l.original_line_number == 30 end).predicted_previous == 12}")
IO.puts("   - A basic all-positive-number line with the same step the whole way gets the correct next prediction: #{Enum.find(o.lines, fn l -> l.original_line_number == 30 end).predicted_next == 100}")
IO.puts("   - A mixed-positive-and-negative line with all same steps line gets correct prev prediction: #{Enum.find(o.lines, fn l -> l.original_line_number == 79 end).predicted_previous == 15}")
IO.puts("   - A mixed-positive-and-negative line with all same steps line gets correct next prediction: #{Enum.find(o.lines, fn l -> l.original_line_number == 79 end).predicted_next == -7}")
IO.puts("")
IO.puts("The sum of all predicted next results is: #{o.sum_of_next_predictions}")

# Got result 1806615041, which was correct.
# Had an error in my logic that was causing each line's prediction to be 2 diffs off, rather than just one. 
# This came from changing the way I was writing the recursive function halfway through and forgetting to
# change the calculation code when I got distracted working on the recursive function itself.

IO.puts("The sum of all predicted previous results is: #{o.sum_of_prev_predictions}")

# Got result 1211, which was correct. 
# That was nice and easy to do because of th estructure I used to build the data, very minimal changes.
# Huzzah for over-engineering?
