defmodule PipeMap.GridSquare do
  alias __MODULE__, as: GS

  defstruct coord: {0,0}, connections: [], symbol: "", distance_from_start: -1, enclosed: nil, which_side_of_path: nil

  def parse({x, y}=coord, map_char) do
    case map_char do
      "S" -> %GS{coord: coord, symbol: map_char}
      "." -> %GS{coord: coord, symbol: map_char}
      "|" -> %GS{coord: coord, symbol: map_char, connections: [{x, y+1}, {x, y-1}]}
      "-" -> %GS{coord: coord, symbol: map_char, connections: [{x+1, y}, {x-1, y}]}
      "L" -> %GS{coord: coord, symbol: map_char, connections: [{x, y-1}, {x+1, y}]}
      "J" -> %GS{coord: coord, symbol: map_char, connections: [{x, y-1}, {x-1, y}]}
      "7" -> %GS{coord: coord, symbol: map_char, connections: [{x-1, y}, {x, y+1}]}
      "F" -> %GS{coord: coord, symbol: map_char, connections: [{x+1, y}, {x, y+1}]}
    end
  end

  def as_string(%GS{}=gs) do
    cond do
      gs.symbol == "S" -> "S"
      gs.distance_from_start > -1 -> "*"
      gs.enclosed -> "E"
      gs.which_side_of_path ->
        case gs.which_side_of_path do
          :left -> "l"
          :right -> "r"
        end
      true -> gs.symbol
    end
  end

  def adjacent_coords(%GS{coord: coord}), do: adjacent_coords(coord)
  def adjacent_coords({sx, sy}) do
    [
      {sx-1, sy-1}, {sx, sy-1}, {sx+1, sy-1},
      {sx-1, sy},               {sx+1, sy},
      {sx-1, sy+1}, {sx, sy+1}, {sx+1, sy+1}
    ]
  end

  def labeled_adjacent_cardinal_coords(%GS{coord: coord}), do: labeled_adjacent_cardinal_coords(coord)
  def labeled_adjacent_cardinal_coords({sx, sy}) do
   %{ 
                      {sx, sy-1} => :n, 
     {sx-1, sy} => :w,                 {sx+1, sy} => :e,
                      {sx, sy+1} => :s,
    }
  end

  @doc """
  When given a gridsquare, finds all adjacent coords that are *NOT* part of the path.

  If given a coordinate that is not part of the path returns an empty list.
  """
  def adjacent_non_path_coords(gs, pipe_map) do
    if gs.distance_from_start < 0 do
      adjacent_coords(gs)
      |> Enum.reject(fn c -> pipe_map[c].distance_from_start > -1 end)
    else
      []
    end
  end

end

defmodule PipeMap do
  #@behaviour Access
  alias PipeMap.GridSquare, as: GS

  defstruct map: %{},
    map_max_x: 0,
    map_max_y: 0,
    starting_coord: {nil, nil},
    furthest_point_distance: -1,
    enclosed_tiles_count: -1

  # Maybe implement Access for PipeMap to pass pipe_map[{x,y}] through to
  # the map inside the struct?
  #def fetch(pm, key) do
    #case Map.get(pm.map, key) do
      #value -> {:ok, value}
      #false -> :error
    #end
  #end


  def from_file(filename) do
    File.stream!(filename)
    |> Enum.with_index()
    |> Enum.map(&Task.async(PipeMap, :parse_line, [&1]))
    |> Enum.map(&Task.await/1)
    |> List.flatten()
    |> Enum.into(%{}, fn gs -> {gs.coord, gs} end)
    |> calc_starting_gs_connections()
    |> calc_map_path_distances_and_build_pipemap_struct()
    |> calc_grid_size()
    |> tap(fn pm -> IO.puts("SANITY CHECK: Grid size is #{pm.map_max_x} by #{pm.map_max_y}, and thus should contain #{pm.map_max_x*pm.map_max_y} items. (Actually contains: #{Enum.count(pm.map)})") end)
    |> calc_left_and_right_of_path()
    |> calc_inside_outside()
    |> fill_enclosed_areas()
    |> calc_enclosed_count()
    |> tap(fn pm -> IO.puts("SANITY CHECK: Final grid:\n#{as_string(pm)}") end)
  end

  def parse_line({line_str, line_num}) do
    line_str
    |> String.trim()
    |> String.graphemes()
    |> Enum.with_index()
    |> Enum.map(fn {char, char_num} -> GS.parse({char_num+1, line_num+1}, char) end)
  end

  def as_string(%PipeMap{}=pm) do
    Enum.reduce(1..pm.map_max_y, "", fn y, y_acc ->
      Enum.reduce(1..pm.map_max_x, y_acc, fn x, x_acc ->
        x_acc <> GS.as_string(pm.map[{x, y}])
      end)
      <> "\n"
    end)
  end

  @doc """
  There should be exactly two connections for each square in the path. The one
  with a distance from start higher than this one is the next step.

  NOTE: This will NOT work on the starting coordinate, since both
  of it's connections are higher. But that doesn't matter in any case because
  the starting coord is a special case in path-walking anyway so we'd never
  use this for that.

  NOTE that this will only work if called on a GridSquare that's actually part
  of the path. Error checking for this here would be nonsensical, you should be
  making sure that you never call it on an incorrect gridsquare.
  """
  def get_next_coord_in_path(%PipeMap{map: m}, {cx, cy}) do
    gs = Map.get(m, {cx, cy})

    Enum.find(gs.connections, fn c ->
      m[c].distance_from_start > gs.distance_from_start
    end)
  end

  defp calc_starting_gs_connections(built_map) do
    {starting_coord, %GS{}} =
      Enum.find(built_map, fn {_k, %{symbol: s}} -> s == "S" end)

    connected_coords = 
      starting_coord
      |> GS.adjacent_coords()
      |> Enum.filter(fn coord ->  # Filter so we only have coordinates which have a connection to the starting coord
        Map.get(built_map, coord, %{connections: []}).connections # Default handles the *literal* edge case where the starting position is on the edge of the grid, so there's nothing in the map for one of it's adjacent coordinates.
        |> Enum.find(fn connecting_coord -> connecting_coord == starting_coord end)
      end)

    # Update the starting coord
    built_map
    |> put_in([starting_coord, Access.key!(:distance_from_start)], 0)
    |> put_in([starting_coord, Access.key!(:connections)], connected_coords)
  end

  defp calc_map_path_distances_and_build_pipemap_struct(built_map) do
    {_key, %GS{} = starting_gs} =
      Enum.find(built_map, fn {_k, %{symbol: s}} -> s == "S" end)

    # We're going to use the total size of the grid as the upper limit of
    # our iterator, since our path is non-cyclic so we know the maximum path
    # possible would be one which covered every single square.
    # We're using reduce_while so that we can abort out of the loop as soon
    # as we find the midpoint.
    {map_with_path_lengths, max_path_length} = 
      1..(Enum.count(built_map))
      |> Enum.reduce_while({built_map, 1, starting_gs.connections}, fn _, {m, step_num, conn_coords} ->
        next_steps_coords = 
          conn_coords
          |> Enum.map(fn c -> m[c].connections end)
          |> List.flatten()
          |> Enum.reject(fn c -> m[c].distance_from_start > -1 end)
        if Enum.count(next_steps_coords) == 0 do
          # This gs *is* the midpoint, because we've reached a point where
          # connections in both directions are marked path components.
          # Now we need to update *this* GS's distance.
          midpoint_coord = 
            conn_coords
            |> Enum.uniq()
            |> Enum.at(0)
          
          {:halt, {put_in(m, [midpoint_coord, Access.key!(:distance_from_start)], step_num), step_num}}
        else
          # Now we have a list of all the next-steps (it should only be 2,
          # but we're making it extensible in case part2 gets silly.)
          next_map = 
            # For every of the "next step" coordinates
            conn_coords
            # Throw away any coordinate which has already had a distance calculated. (That's either the connection back to where we just were, or we've reached the midpoint.)
            |> Enum.reject(fn c -> m[c].distance_from_start > -1 end)
            # Now for each of the next coordinates we update their distance measure with our current distance
            |> Enum.reduce(m, fn coord, acc_m -> put_in(acc_m, [coord, Access.key!(:distance_from_start)], step_num) end)


          {:cont, {next_map, step_num+1, next_steps_coords}}
        end
      end)
    
    %PipeMap{map: map_with_path_lengths, furthest_point_distance: max_path_length, starting_coord: starting_gs.coord}
  end

  defp calc_grid_size(%PipeMap{}=pipe_map) do
    {max_x, max_y} = pipe_map.map
                      |> Enum.reduce({0,0}, fn {{tx, ty}, _}, {ax, ay} -> {max(tx, ax), max(ty, ay)} end)

    %{pipe_map| map_max_x: max_x, map_max_y: max_y}
  end

  defp calc_left_and_right_of_path(%PipeMap{}=pipe_map) do
    # We're going to walk the path, in both directions, and label things that
    # are on either the "Left" or "right" of the path if we walked it from
    # the start all the way around back to the start.
    # This means that as we walk up on one side "left" is actually "right".

    starting_gs = Map.get(pipe_map.map, pipe_map.starting_coord)

    pipe_map
    |> mark_left_and_right_for_path_starting_at(Enum.at(starting_gs.connections, 0), :normal)
    |> mark_left_and_right_for_path_starting_at(Enum.at(starting_gs.connections, 1), :reversed)
  end

  defp mark_left_and_right_for_path_starting_at(%PipeMap{}=pipe_map, {tx, ty}, path_side) do
    # Takes a single square or coord, and the map it comes from, as well as a
    # flag indicating if this is the "normal" side of the path, or the
    # "reversed" side. (Walking the path both directions simultaneously
    # is required to be able to figure out which is "left" and "right".
    # Returns the updated pipe_map
    # NOTE: There is a slight redundancy here in that both directions will
    #       mark the final square. This shouldn't matter though because they'll
    #       both mark it as the same thing.
    next_coord = get_next_coord_in_path(pipe_map, {tx, ty})

    case next_coord do
      nil -> pipe_map
      nxt -> 
        mark_left_and_right_for_path_starting_at(pipe_map, nxt, path_side)
        |> mark_left_right_for_coord({tx, ty}, next_coord, path_side)
    end
  end

  defp mark_left_right_for_coord(%PipeMap{}=pipe_map, cur_coord, next_coord, path_side) do
    next_coord_cardinal =
      GS.labeled_adjacent_cardinal_coords(cur_coord)
      |> Enum.find(fn {cardinal_coord, _cardinal_dir} -> cardinal_coord == next_coord end)
      |> elem(1)

    case pipe_map.map[cur_coord].symbol do
      "S" -> :error
      "." -> :error
      "|" -> case next_coord_cardinal do
        :n -> %{w: :left, e: :right}
        :s -> %{w: :right, e: :left}
        _ -> :error
      end
      "-" -> case next_coord_cardinal do
        :e -> %{n: :left, s: :right}
        :w -> %{n: :right, s: :left}
        _ -> :error
      end
      "L" -> case next_coord_cardinal do
        :e -> %{w: :right, s: :right}
        :n -> %{s: :left, w: :left}
        _ -> :error
      end
      "J" -> case next_coord_cardinal do
        :w -> %{e: :left, s: :left}
        :n -> %{s: :right, e: :right}
        _ -> :error
      end
      "7" -> case next_coord_cardinal do
        :w -> %{e: :right, n: :right}
        :s -> %{n: :left, e: :left}
          _ -> :error
      end
      "F" -> case next_coord_cardinal do
        :e -> %{w: :left, n: :left}
        :s -> %{n: :right, w: :right}
        _ -> :error
      end
    end
    |> (fn m -> 
      case path_side do
        :normal -> m
        :reversed -> Enum.map(m, fn {dir, side} -> 
          new_side = case side do
            :left -> :right
            :right -> :left
          end
          {dir, new_side}
        end)
      end
    end).()
    |> Enum.reduce(pipe_map.map, fn {cardinal_dir, has_side}, acc -> 
      # My collegue here, who is too cunning to be understood....
      # We're going to reduce across the struct piped in from above containing
      # a map of {:w, :left} type things, which we will use to get the the
      # coord from that cardinal direction, and then update it in the pipe_map
      # with the correct side
      GS.labeled_adjacent_cardinal_coords(cur_coord)
      |> Enum.find(fn {_adjacent_coord, adjacent_cardinal_side} -> adjacent_cardinal_side == cardinal_dir end)
      |> elem(0)
      |> then(fn c -> Map.get(acc, c) end)
      |> case do
        nil -> acc
        gs -> 
          # Don't update "side" value if the adjacent GS is part of the path
          case gs.distance_from_start do
            d when d < 0 -> put_in(acc, [gs.coord, Access.key!(:which_side_of_path)], has_side)
            _ -> acc
          end
      end
    end)
    |> then(fn m -> %{pipe_map| map: m} end)
  end


  defp calc_inside_outside(%PipeMap{}=pipe_map) do
    # Now that everything is marked we just need to figure out which "side" is
    # touching the edge, or has an path to the edge that is not blocked by the
    # path.
    # Whichever side *ISN'T* touching or having a path to the edge can be
    # marked as "enclosed".

    max_x = pipe_map.map_max_x
    max_y = pipe_map.map_max_y

    sides_touching_edges =
      pipe_map.map
      |> Enum.filter(fn
        {{1, _y}, _gs} -> true
        {{_x, 1}, _gs} -> true
        {{^max_x, _y}, _gs} -> true
        {{_x, ^max_y}, _gs} -> true
        _ -> false
      end)
      |> Enum.map(fn {_coord, gs} -> gs.which_side_of_path end)
      |> Enum.uniq()
      |> Enum.reject(fn x -> x == nil end) # many of the squares won't have a "Side"; throw them out.

    if Enum.count(sides_touching_edges) == 1 do
      # This is the good case where there is exactly one "side"
      # (Really it can be zero or one) touching an outer edge.
      # This automatically implies that this "side" is "outside" the path, and
      # we can skip straight on to the next step.
      case Enum.at(sides_touching_edges, 0) do
        :left -> PipeMap.mark_side_as_enclosed({pipe_map, :right})
        :right -> PipeMap.mark_side_as_enclosed({pipe_map, :left})
      end
    else
      # So this is the bad case. Since no side of the path is directly adjacent
      # to the edge, now we're going to have to start searching for a side that
      # can find a way to walk to the outside without crossing the path.
      # NOTE: This will have a degenerate case where the **ENTIRE** "outside"
      # side of the path will only be accessible by squeezing in between pipes.
      # After a little testing, our input doesn't seem to do this, so.... Not 
      # going to bother dealing with that. Bad programmer.
      pipe_map.map
      |> Enum.group_by(fn {_coord, gs} -> gs.which_side_of_path end)
      |> Enum.reject(fn {s, _gss} -> s == nil end)
      |> Enum.reduce_while(%{}, fn {side, grid_squares}, _acc ->
        # Now we'll walk each grid square for this side and determine if it can
        # reach the edge with a straight cardinal walk in any direction without
        # bumping into a path square.
        Enum.reduce_while(grid_squares, nil, fn {_coord, gs}, _acc ->
          case can_walk_any_cardinal_to_edge(pipe_map, gs) do
            true -> {:halt, true}
            false -> {:cont, false}
          end
        end)
        |> case do
          #TODO: We should be returning the OPPOSITE side here to be marked as enclosed
          true ->
            case side do
              :left -> {:halt, {pipe_map, :right}}
              :right -> {:halt, {pipe_map, :left}}
            end
          false -> {:cont, :error}
        end
      end)
      |> case do
        :error -> 
          IO.puts("ERROR!!!!!! This should *NOT* happen, but it seems that neither side could walk to an edge!")
          pipe_map
        res -> PipeMap.mark_side_as_enclosed(res)
      end
    end
  end

  defp fill_enclosed_areas(%PipeMap{}=pipe_map) do
    # We've marked which side is "enclosed", which means that if there are any
    # large blocks that are entirely enclosed, we can fill their "enclosed"
    # flag. We know that any non-path grid square they are touching that's not
    # already marked as enclosed can also be marked as enclosed.

    new_map = 
      pipe_map.map
      |> Enum.filter(fn {_coord, gs} -> gs.enclosed end)
      |> Enum.reduce(pipe_map.map, fn {_coord, gs}, acc_map ->
        recursive_fill_neighbors_as_enclosed(gs, acc_map)
      end)

    %{pipe_map| map: new_map}
  end

  defp recursive_fill_neighbors_as_enclosed(%GS{}=gs, map) do
    # Takes a grid square that we know is enclosed and fill in any non-path,
    # neighbors as also being enclosed, recursively of course.
    
    GS.adjacent_coords(gs)
    |> Enum.map(fn c -> Map.get(map, c) end)
    |> Enum.reject(fn gs -> gs.distance_from_start > -1 or gs.enclosed end)
    |> Enum.reduce(map, fn gs, acc_map ->
      recursive_fill_neighbors_as_enclosed(gs, put_in(acc_map, [gs.coord, Access.key!(:enclosed)], true))
    end)
  end

  defp can_walk_any_cardinal_to_edge(%PipeMap{}=pipe_map, %GS{}=gs) do
    [{-1, 0}, {1, 0}, {0, -1}, {0, 1}]
    |> Enum.map(&Task.async(PipeMap, :can_walk_cardinal_to_edge, [pipe_map, gs.coord, &1]))
    |> Enum.map(&Task.await/1)
    |> Enum.find(false, fn d -> d == true end)
  end

  def can_walk_cardinal_to_edge(%PipeMap{}=pipe_map, {x, y}, {dx, dy}=cardinal_operator) do
    next_x = x+dx
    next_y = y+dy

    if next_x >= pipe_map.map_max_x
      or next_x <= 0
      or next_y >= pipe_map.map_max_y
      or next_y <= 0 do
      true
    else
      next_gs = pipe_map.map[{next_x, next_y}]
      if next_gs.symbol == S or next_gs.distance_from_start >= 0 do
        false
      else
        can_walk_cardinal_to_edge(pipe_map, {next_x, next_y}, cardinal_operator)
      end
    end
  end

  def mark_side_as_enclosed({%PipeMap{}=pipe_map, enclosed_side}) do
    Enum.reduce(pipe_map.map, %{}, fn {coord, gs}, acc ->
      case gs.which_side_of_path do
        ^enclosed_side -> %{gs| enclosed: true}
        _ -> gs
      end
      |> then(fn gs -> Map.put(acc, coord, gs) end)
    end)
    |> then(fn m -> %{pipe_map| map: m} end)
  end

  #defp old_calc_inside_outside(%PipeMap{}=pipe_map) do
    ## Now that we have the squares immediately left and right, we can determine
    ## which is "inside" and "outside" the path. 
    ## The first thing we need to do is fill in any blocks that can be inferred.
    ## For example, after walking the left-rigth values of the path we get the
    ## following:
    ##
    ## .....
    ## LLLLL
    ## 12345
    ## 2RRR6
    ## 3R.R7
    ## 4RRR8
    ## 56789
    ## The very center square will not be marked as R. Because it is adjacent
    ## to only R squares though we can infer that it is indeed an "R" square,
    ## and the upper row of L squares can also be inferred as L.
    ##
    ## We need to fill in these areas so that we can determine if any zones
    ## touch the edge. In this case, we'll be able to infer that "R" is the
    ## inside and "L" is the outside because an "L" block touches the outer
    ## edge of the grid.

    #Enum.reduce(1..Enum.count(pipe_map.map), pipe_map.map, fn {_coord, gs}, acc_map -> 
      #if gs.which_side_of_path do
        #recurse_fill_side_info(acc_map, gs, gs.which_side_of_path)
      #else
        #acc_map
      #end
    #end) # Now we have the map with the zones filled.
    #|> determine_which_side_is_external()
    #|> mark_inside_outside()
    #|> then(fn m -> %{pipe_map| map: m} end)
  #end

  #defp recurse_fill_side_info(%PipeMap{map: map}=pipe_map, %GS{}=gs, path_side) do
    #if gs.which_side_of_path or gs.distance_from_start > -1 do
      ## We've hit either an already filled square, or the path itself, so we've
      ## found the edges of the zone we're filling, so we're done.
      #pipe_map
    #else
      #GS.adjacent_coords(gs)
      #|> Enum.map(fn c -> Map.get(map, c) end)
      #|> Enum.reject(fn gs -> gs == nil end)
      #|> Enum.reduce(map, fn next_gs, acc_map ->
        #recurse_fill_side_info(acc_map, next_gs, path_side)
      #end)
      #|> then(fn pm -> put_in(pm, [Access.key!(:map), gs.coord, Access.key!(:which_side_of_path)], path_side) end)
    #end
  #end

  ## Pipe-specific function used to determine which "side" of the path is external.
  ## Returns a {pipe_map, side} tuple for use in the next function in the pipe
  ## TODO: Change this function to use the take-2 method.
  #defp determine_which_side_is_external(%PipeMap{}=pipe_map) do
    #max_x = pipe_map.map_max_x
    #max_y = pipe_map.map_max_y

    #sides_touching_edges =
      #pipe_map.map
      #|> Enum.filter(fn
        #{{1, _y}, _gs} -> true
        #{{_x, 1}, _gs} -> true
        #{{^max_x, _y}, _gs} -> true
        #{{_x, ^max_y}, _gs} -> true
        #_ -> false
      #end)
      #|> Enum.map(fn {_coord, gs} -> gs.side end)
      #|> Enum.uniq()
      #|> Enum.reject(fn x -> x == nil end) # many of the squares won't have a "Side"; throw them out.

    #if Enum.count(sides_touching_edges) == 1 do
      ## This is the good case where there is exactly one "side"
      ## (Really it can be zero or one) touching an outer edge.
      ## This automatically implies that this "side" is "outside" the path, and
      ## we can skip straight on to the next step.
      #{pipe_map, Enum.at(sides_touching_edges, 0)}
    #else
      ## So this is the bad case. Since no side of the path is directly adjacent
      ## to the edge, now we're going to have to start searching for a side that
      ## can find a way to walk to the outside without crossing the path.
      ## NOTE: This will have a degenerate case where the **ENTIRE** "outside"
      ## side of the path will only be accessible by squeezing in between pipes.
      
      ## TODO: WALK TO EDGES
      #{pipe_map, :error}
    #end

  #end

  #defp mark_inside_outside({%PipeMap{map: m}=pipe_map, side}) do
    ## update the pipe_map where all gridsquares
    ## equivalent to side get marked as enclosed: true
    #%{pipe_map| map: Enum.reduce(m, %{}, fn {c, gs}, acc_map -> 
      #this_grid = case gs.which_side_of_path == side do
        #true -> {c, %{gs| enclosed: true}}
        #false -> {c, gs}
      #end
      #Map.put(acc_map, c, this_grid)
    #end)}
  #end

  defp calc_enclosed_count(%PipeMap{}=pipe_map) do
    %{pipe_map| enclosed_tiles_count: Enum.filter(pipe_map.map, fn {_c, gs} -> gs.enclosed end) |> Enum.count()}
  end

end

IO.puts("Parsing test input...")
tm = PipeMap.from_file("test_input")
IO.puts("TEST: Path length = #{tm.furthest_point_distance}, (should be 80), and begins at #{inspect tm.starting_coord}")
IO.puts("TEST: Tiles enclosed by loop = #{tm.enclosed_tiles_count} (Should be 10)")
IO.puts("TEST: One of the first steps is: #{inspect tm.map[{4,1}]}")
IO.puts("TEST: Midpoint is: #{inspect tm.map[{16,8}]}")
Enum.filter(tm.map, fn {_k, v} -> v.distance_from_start == tm.furthest_point_distance-1 end)
|> Enum.count()
|> IO.inspect(label: "TEST: Number of nodes immediately prior to furthest point")
IO.puts("")

IO.puts("Parsing Pipe Map...")
m = PipeMap.from_file("input")
{{_start_x, _start_y}, starting_gs} = Enum.find(m.map, fn {_k, %{symbol: s}} -> s == "S" end)
IO.puts("Starting from position #{inspect m.starting_coord} (which is connected to: #{inspect starting_gs.connections})...")
IO.puts("Steps to get furthest down path: #{m.furthest_point_distance}")
IO.puts("Tiles enclosed by loop: #{m.enclosed_tiles_count}")


# Got 6812, which was correct! 
# Part 2 got 527, eventually, which was correct! 

# This project was cursed by off-by-one errors, one of the classic blunders...
# A lot of this came from iterating maps, and not doing Enum.count()-1, but
# I should've thought more clearly about how to set up the indexing in the
# beginning too. Perhaps written an abstraction layer to use?
#
# I also learned that I need to get better at using step debugging in Elixir...
