defmodule Galaxy do
  @moduledoc """
  Represents a single galaxy, somewhere in space.

  Tracks the galaxy based on it's coordinates in our GalaxyMap picture.
  """
  defstruct id: -1, x: -1, y: -1
end

defmodule GalaxyMap do
  @moduledoc """
  Represents a map of the galaxy. Because space, you see, is big, really really
  big, what we actually map is simply a list of galaxies, ignoring the spaces
  between.

  Our GalaxyMap is actually based on a picture of a certain amount of the sky.
  Because of this we track the galaxy coordinates based on their position on
  our picture. Map coordinates are 1-indexed, with the top-left position being
  1,1, the position to it's right being 1,2, etc....
  """
  defstruct width: -1, height: -1, galaxies: [], pairs: %{}

  @doc """
  Take a filename and return a %GalaxyMap{} representing it.
  """
  def from_file(filename) do
    {accumulated_galaxies, width, height} = 
      File.stream!(filename)
      |> Enum.with_index(1)
      |> Enum.reduce({[], -1, -1}, fn {line_string, line_number}, {acc, acc_max_x, acc_max_y} ->
        {accumulated_line, line_length} = 
          line_string
          |> String.trim()
          |> String.graphemes()
          |> Enum.with_index(1)
          |> Enum.reduce({acc, -1}, fn {char, char_number}, {line_acc, line_len_acc} ->
            case char do
              c when c == "#" -> {line_acc ++ [%Galaxy{x: char_number, y: line_number}], max(char_number, line_len_acc)}
              _ -> {line_acc, max(char_number, line_len_acc)}
            end
          end)

        {accumulated_line, max(acc_max_x, line_length), max(acc_max_y, line_number)}
      end)

    indexed_galaxies = 
      accumulated_galaxies
      |> Enum.with_index(1)
      |> Enum.map(fn {galaxy, idx} -> %{galaxy| id: idx} end)

    %GalaxyMap{galaxies: indexed_galaxies, width: width, height: height}
  end

  @doc """
  Convert a map back to it's string form.
  """
  def as_string(%GalaxyMap{}=gm, debug \\ true, show_path_for_pair \\ nil) do
    line_grouped_galaxies = Enum.group_by(gm.galaxies, &(&1.y))

    Enum.reduce(1..gm.height, "", fn line_number, acc ->
      this_line_galaxies = Map.get(line_grouped_galaxies, line_number, []) # Handle the empty lines with a default

      this_line = 
        1..gm.width
        |> Enum.map(fn _tx -> "." end)
        |> (fn line_list -> 
          Enum.reduce(this_line_galaxies, line_list, fn g, line_acc ->
            List.replace_at(line_acc, g.x-1, "#") # NOTE: We have to fix the off-by-one for Enum.at being zero-indexed
          end)
        end).()
        |> Enum.join()

      if debug do
        acc <> this_line <> " #{(Integer.to_string(line_number) |> String.pad_leading(3, "0"))}, IDs: #{inspect(Enum.map(this_line_galaxies, &(&1.id)), charlists: :as_list)}" <> "\n"
      else
        acc <> this_line <> "\n"
      end
    end)
  end

  @doc """
  Take a GalaxyMap and return a new one that's been expanded.

  Expansion is defined as taking any line that has zero galaxies, and adding a new line after it.
  """
  def expand(%GalaxyMap{}=gm, amount \\ 1) do
    filled_lines = 
      gm.galaxies
      |> Enum.map(&(&1.y))
      |> Enum.sort()
      |> Enum.uniq()
    all_lines = 1..Enum.at(filled_lines, -1) |> Range.to_list()
    # We want empty lines, but we're going to pre-calculate the fact that each
    # of these lines is going to increase the line number of all following
    # lines. For example, if our filled lines are:
    # [1,3,5]
    # then our empty lines are [2,4]
    # but when we "expand" by adding an extra empty line at 2, our new list is
    # [1,{2},{3},4,6] (where 2 and 3 are empty)
    # which means our last remaining empty line is 5 instead of 4
    # We can solve this by pre-computing that for every element in the empty
    # lines list we need to increase it's value by it's zero-indexed position
    # in the list, such that the following filled-lines list:
    # [1,3,5,7,9]
    # will result in an empty list:
    # [2,4,6,8]
    # the pre-computed version of that is:
    # [2+0, 4+1, 6+2, 8+3], or [2,5,8,11]
    # and a final list of:
    # [1,{2},{3},4,{5},{6},7,{8},{9},10,{11},{12},13]
    # Which is the correct expansion.
    empty_lines =
      all_lines -- filled_lines
      |> Enum.sort()
      |> Enum.with_index()
      |> Enum.map(fn {val, idx} -> val+(idx*amount) end)
    IO.puts(" - Galaxy Expansion found #{Enum.count(empty_lines)} lines to expand. (#{inspect empty_lines})")

    filled_cols = 
      gm.galaxies
      |> Enum.map(&(&1.x))
      |> Enum.sort()
      |> Enum.uniq()
    all_cols = 1..gm.width |> Range.to_list()
    # Once again, we're going to do the pre-compute thing so as we expand
    # the col_ids that are getting expanded are based on having already
    # expanded all previous cols in the list.
    empty_cols =
      all_cols -- filled_cols
      |> Enum.sort()
      |> Enum.with_index()
      |> Enum.map(fn {val, idx} -> val+(idx*amount) end)
    IO.puts(" - Galaxy Expansion found #{Enum.count(empty_cols)} lines to expand. (#{inspect empty_cols})")

    # Iterate across this list of pre-computed empty lines, updating the
    # galaxies list, 
    with_expanded_rows = Enum.reduce(empty_lines, gm.galaxies, fn empty_line_num, acc_list ->
      # Update every galaxy after this empty line.
      Enum.reduce(acc_list, [], fn galaxy, update_acc ->
        update_acc
        ++ case galaxy.y > empty_line_num do
          true -> [%{galaxy| y: galaxy.y+amount}]
          false -> [galaxy]
        end
      end)
    end)

    # Iterate across the pre-computer list of empty cols and push every galaxy
    # with an X coord higher than this empty col up by one.
    updated_galaxies = Enum.reduce(empty_cols, with_expanded_rows, fn empty_col_num, acc_list ->
      # Update every galaxy's col position that's higher than this one.
      Enum.reduce(acc_list, [], fn galaxy, col_update_acc ->
        col_update_acc
        ++ case galaxy.x > empty_col_num do
          true -> [%{galaxy | x: galaxy.x+amount}]
          false -> [galaxy]
        end
      end)
    end)

    %{gm|
      galaxies: updated_galaxies,
      height: Enum.map(updated_galaxies, &(&1.y)) |> Enum.sort() |> Enum.at(-1),
      width: Enum.map(updated_galaxies, &(&1.x)) |> Enum.sort() |> Enum.at(-1)
    }
  end

  @doc """
  Accepts a GalaxyMap, and calculates/populates it's "pairs".

  The resulting dict will look like:
  %{
    {1, 2} => %{shortest_path_step_distance: 3},
    {1, 3} => %{shortest_path_step_distance: 2},
    {2, 3} => %{shortest_path_step_distance: 1}
  }

  Note that each pair is bi-directional. It's ID is always
  {lower_galaxy_id, higher_galaxy_id}. 
  Use &id_for_pair/2 to ensure generation of a correct ID
  when attempting to look up pairs in the GalaxyMap.pairs map.
  """
  def populate_shortest_path_pairs(%GalaxyMap{}=gm) do
    gm.galaxies
    |> Enum.reduce(%{}, fn galaxy, acc ->
      gm.galaxies
      |> Enum.reject(fn iter_galaxy -> iter_galaxy.id == galaxy.id end)
      |> Enum.map(fn iter_galaxy -> id_for_pair(iter_galaxy, galaxy) end)
      |> Enum.into(acc, fn {g1_id, g2_id}=pair_id ->
        g1 = Enum.find(gm.galaxies, fn g -> g.id ==  g1_id end)
        g2 = Enum.find(gm.galaxies, fn g -> g.id ==  g2_id end)
        {pair_id, %{shortest_path_step_distance: step_distance_between(g1, g2)}}
      end)
    end)
    |> then(fn pairs_map -> %{gm| pairs: pairs_map} end)
  end

  @doc """
  Accepts a GalaxyMap and returns the sum of of all the shortest paths.
  Requires a GalaxyMap that has had the shortest path pairs populated.
  """
  def sum_of_shortest_paths(%GalaxyMap{}=gm) do
    gm.pairs
    |> Enum.map(fn {_pair_id, pair_map} -> pair_map.shortest_path_step_distance end)
    |> Enum.sum()
  end

  @doc """
  This function creates a consistent {1, 2} id tuple, independent of the
  order of the galaxys in the arguments.
  This consistent "id" format for a pair allows us to qucikly determine
  if a path between B and A already exists as path A->B.
  """
  def id_for_pair(galaxy1, galaxy2) do
    {min(galaxy1.id, galaxy2.id), max(galaxy1.id, galaxy2.id)}
  end


  @doc """
  Calculates the shortest number of steps to get from galaxy g1 to galaxy g2.
  The diagonal path step distance is actually the same as making a path of
  2 perpendicular lines.
  (2 down, 2 over, is the same as 1 down 1 over 1 down 1 over)
  This works because we're counting *steps* not distance. No trig.
  """
  def step_distance_between(%Galaxy{}=g1, %Galaxy{}=g2) do
    (max(g1.x, g2.x)-min(g1.x, g2.x)) + (max(g1.y, g2.y)-min(g1.y, g2.y))
  end

  @doc """
  Calculate, mathematically, the expected number of pairs for the given
  GalaxyMap and it's list of galaxies
  """
  def predicted_number_of_pairs(%GalaxyMap{}=gm) do
    n = Enum.count(gm.galaxies)
    (n * (n-1)) / 2
  end

  def print_debug_pair(%GalaxyMap{}=gm, pair_id) do
    IO.puts("The pair #{inspect(pair_id)} is: #{gm.pairs[pair_id] |> inspect()}. #{Enum.find(gm.galaxies, fn x -> x.id == elem(pair_id, 0) end) |> inspect()}, #{Enum.find(gm.galaxies, fn x -> x.id == elem(pair_id, 1) end) |> inspect()}")
  end
end


_gm =
  GalaxyMap.from_file("input")
  |> tap(fn xgm -> IO.puts("Inital parsing gives us #{xgm.width} by #{xgm.height} picture, with #{Enum.count(xgm.galaxies)} galaxies.") end)
  |> GalaxyMap.expand()
  |> tap(fn xgm -> IO.puts("After expansion: #{Enum.count(xgm.galaxies)} galaxies, on a picture that is #{xgm.width} by #{xgm.height}") end)
  |> GalaxyMap.populate_shortest_path_pairs()
  |> tap(fn xgm -> IO.puts("Galaxy contains #{Enum.count(xgm.pairs)} pairs. (Mathematically there should be #{GalaxyMap.predicted_number_of_pairs(xgm)})") end)
  |> tap(fn xgm -> IO.puts(GalaxyMap.as_string(xgm)) end)
  |> tap(fn xgm -> GalaxyMap.print_debug_pair(xgm, {1, 12}) end)
  |> tap(fn xgm -> IO.puts("The sum of all the shortest paths is: #{GalaxyMap.sum_of_shortest_paths(xgm)}") end)

# Part 1 got 9177603, which was right! 
# No major issues in Part1, moving everything to 1-indexed helped, except for
# the *ONE* place where we used Enum.at() and had to compensate. That took a
# little bit to figure out. Sigh.

IO.puts("-------------------------\nDoing Mega Expansion Part2!")
GalaxyMap.from_file("input")
|> GalaxyMap.expand(999_999)
|> tap(fn xgm -> IO.puts("After expansion: #{Enum.count(xgm.galaxies)} galaxies, on a picture that is #{xgm.width} by #{xgm.height}") end)
|> GalaxyMap.populate_shortest_path_pairs()
|> tap(fn xgm -> IO.puts("Galaxy contains #{Enum.count(xgm.pairs)} pairs. (Mathematically there should be #{GalaxyMap.predicted_number_of_pairs(xgm)})") end)
|> tap(fn xgm -> IO.puts("The sum of all the shortest paths is: #{GalaxyMap.sum_of_shortest_paths(xgm)}") end)

# Part 2 was 632003913611.
# It was super easy because of the way I did the calculation/data-structure
# and didn't even take any extra time.
# It took me 2 attempts because I off-by-one'd it again and expanded by
# 1_000_000 as requested, but because my function actualy increases by
# the amount + the original line I got a value that was too high.
