require IEx

defmodule ConditionList do
  @moduledoc """
  Represents a list of condition strings.

  Note that ordering of the conditions is not necessarily maintained.
  """
  defstruct conditions: []

  @doc """
  Read a file from a filename, and parse each line as a unique Condition,
  returning a ConditionList struct.
  """
  def from_file(filename \\ "input") do
    File.stream!(filename)
    |> Enum.map(&Condition.parse/1)
    #|> Enum.map(&Task.async(Condition, :parse, [&1]))
    #|> Enum.map(&Task.await/1)
    |> then(fn conditions -> %ConditionList{conditions: conditions} end)
  end

  @doc """
  Each Condition in a ConditionList has some number >=0 of possible ways
  to arrange the good/bad springs that will fulfill both the malformed
  record string, and the contiguous grouping string.

  This function will sum the possible number of arrangements of all conditions
  in a given ConditionList.
  """
  def sum_of_arrangement_counts(%ConditionList{conditions: conditions}) do
    conditions
    |> Enum.map(fn c -> Enum.count(c.arrangements) end)
    |> Enum.sum()
  end
end


defmodule Condition do
  @moduledoc """
  A condition is a record of a row of springs and their state.
  ? is unknown
  . is operation
  # is damaged.

  And it is followed by a numerical list which represents groups of
  contiguous broken springs.
  So ???.### 1,1,3 MUST be #.#.###
  """
  defstruct arrangements: [], original_string: ""

  @doc """
  Accepts a single line from a file, which contains a malformed record string,
  followed by a space, followed by a comma-separated list representing groups
  of contiguous broken springs.
  """
  def parse(string) do
    string
    |> String.trim()
    |> String.split(" ")
    |> then(fn [record_string, groups] ->
      groups
      |> String.split(",")
      |> Enum.map(&Integer.parse/1)
      |> Enum.map(fn {x, _rem} -> x end)
      |> then(fn integer_list ->
        {record_string, integer_list}
      end)
    end)
    |> build_arrangement_possibilities()

    |> then(fn arrangements ->
      %Condition{original_string: string, arrangements: arrangements}
    end)
  end

  # NOTE: this will return some bad permutations at the top-level, which is
  # can't be filtered here without breaking the recursion.
  # The top-level call of this function should immediately filter out any
  # permutations that still have any integer 
  defp build_arrangement_possibilities({record_string, integer_list}) when record_string == "", do: {[], integer_list}
  defp build_arrangement_possibilities({record_string, integer_list}) do
    [_all, leading_good, this_block, remaining] = Regex.run(~r/(\.*)([#,\?]*)(.*)/, record_string)

    this_block_contains_bad = String.contains?(this_block, "#")

    {tail_possibilities, integer_list_after_tail} =
      case this_block_contains_bad do
        true ->
          # Because this block contains bad we know that it *NEEDS* at least
          # the head of the bad block list, so we only pass the tail further
          # down the recursion. 
          build_arrangement_possibilities(this_block, tl(integer_list))
        false ->
          build_arrangement_possibilities(this_block, integer_list)
      end

    permutations_for_this_block =
      case this_block_contains_bad do
        true ->
          # Because this block contains bad we know that it *NEEDS* at least
          # the head of the bad block list, but we won't know what the other
          # possible bad block group sizes are until after tail has returned.
          permutations_for_this_block =
            calc_permutations_for(this_block, hd(integer_list) ++ integer_list_after_tail)
        false -> 
          permutations_for_this_block =
            calc_permutations_for(this_block, integer_list_after_tail)
      end

    # TODO: If fold all the way back and there are still leftover bad blocks
    # then the whole permutation was invalid.
    # Example:
    #   .?#?.??.?# 1,1,1
    # If we don't return an error in the 2nd block then every single
    # permutation that doesn't contain an error in the 2nd block is
    # invalid, but we won't know this until we return the final string.

    permutations_for_this_block
    |> Enum.map(fn this_block_permutation ->
      Enum.reduce(tail_possibilities, [], fn tail_permutation, acc ->
        acc ++ [leading_good <> this_block_permutation <> tail_permutation]
      end
    end)
    |> List.flatten()
    |> then(fn flattened_list ->
      {flattened_list, remaining_integer_list_after_tail_and_this_block}
    end)
  end

  # At this point we know we're getting a string that contains only ? and #,
  # and we need to return all possible permutations of the string which satisfy
  # at least the final element in the bad_block_grouping_list, but possibly
  # more elements, starting from the last and working towards the first.
  # Each permutation returned needs to be returned as a tuple with the
  # permutation string, as well as the remaining bad_block_group_list elements
  # that were *NOT* used in creating it. 
  # Permutations passed in MAY contain # or ?
  # Permutations passed in WILL NOT contain .
  # Permutations returned MUST NOT contain ?
  defp calc_permutations_for(s, il) when s == "", do: [{"", il}]
  defp calc_permutations_for(s, il) when il == [], do: [{String.replace("?", "."), il}]
  defp calc_permutations_for(string, bad_block_grouping_list) do
    block_must_contain_bad = String.contains?(string, "#")
    # We're going to reduce across the possible list of contiguous bad block
    # sizes, starting at the end, until we reach the point where there are too
    # many bad blocks to fit in the available space, which is how we'll know we
    # got all of the possible permutations.
    bad_block_grouping_list
    |> Enum.reverse()
    |> Enum.reduce_while(%{permutations: [], used_groups: []}, fn bad_block_size, acc -> 
      required_space_for_bad = [bad_block_size] ++ acc[:used_groups] |> Enum.intersperse(1) |> Enum.sum()
      if length(string) < required_space_for_bad do
        # If there aren't enough characters in this block to even do a solid
        # string of bad characters matching the final number in the list,
        # then this *must* be a string of fully good, and we can't possibly
        # need to use any further digits off the contiguous block list.
        if block_must_contain_bad, do: IO.puts("ERROR: We found an error character in string block (#{inspect string}) that couldn't fit the required number of error characters. (#{required_space_for_bad})")
        {:halt,
          %{acc|
            permutations: map[:permutations] ++ [{String.replace(string, "?", "."), bad_block_grouping_list}]
          }
        }
      else
        number_of_possible_bad_blocks = 

        {:cont,
          %{acc|
            permutations: ,
            used_groups: 
          }
        }
      end
    end)
  end
end

cl = ConditionList.from_file()
IO.puts("Our condition list contains #{Enum.count(cl.conditions)} different conditions.")
IO.puts("The first condition is: #{inspect(Enum.at(cl.conditions, 0), limit: :infinity)}")
IO.puts("The sum of the possible arrangements of all conditions is: #{ConditionList.sum_of_arrangement_counts(cl)}")

